@@ -1,6 +1,7 @@
 import torch
 from torch import nn, einsum
 import torch.nn.functional as F
+import torch.distributed as distributed
 from torch.cuda.amp import autocast
 
 from einops import rearrange, repeat
@@ -12,9 +13,8 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
-@contextmanager
-def null_context():
-    yield
+def noop(*args, **kwargs):
+    pass
 
 def l2norm(t):
     return F.normalize(t, p = 2, dim = -1)
@@ -85,7 +85,9 @@ class EuclideanCodebook(nn.Module):
         kmeans_iters = 10,
         decay = 0.8,
         eps = 1e-5,
-        threshold_ema_dead_code = 2
+        threshold_ema_dead_code = 2,
+        use_ddp = False,
+        learnable_codebook = False
     ):
         super().__init__()
         self.decay = decay
@@ -97,11 +99,18 @@ class EuclideanCodebook(nn.Module):
         self.eps = eps
         self.threshold_ema_dead_code = threshold_ema_dead_code
 
+        self.all_reduce_fn = distributed.all_reduce if use_ddp else noop
+
         self.register_buffer('initted', torch.Tensor([not kmeans_init]))
         self.register_buffer('cluster_size', torch.zeros(codebook_size))
-        self.register_buffer('embed', embed)
         self.register_buffer('embed_avg', embed.clone())
 
+        self.learnable_codebook = learnable_codebook
+        if learnable_codebook:
+            self.embed = nn.Parameter(embed)
+        else:
+            self.register_buffer('embed', embed)
+
     @torch.jit.ignore
     def init_embed_(self, data):
         if self.initted:
@@ -131,13 +140,16 @@ class EuclideanCodebook(nn.Module):
         batch_samples = rearrange(batch_samples, '... d -> (...) d')
         self.replace(batch_samples, mask = expired_codes)
 
+    @autocast(enabled = False)
     def forward(self, x):
         shape, dtype = x.shape, x.dtype
         flatten = rearrange(x, '... d -> (...) d')
-        embed = self.embed.t()
 
         self.init_embed_(flatten)
 
+        embed = self.embed if not self.learnable_codebook else self.embed.detach()
+        embed = self.embed.t()
+
         dist = -(
             flatten.pow(2).sum(1, keepdim=True)
             - 2 * flatten @ embed
@@ -150,8 +162,14 @@ class EuclideanCodebook(nn.Module):
         quantize = F.embedding(embed_ind, self.embed)
 
         if self.training:
-            ema_inplace(self.cluster_size, embed_onehot.sum(0), self.decay)
+            cluster_size = embed_onehot.sum(0)
+            self.all_reduce_fn(cluster_size)
+
+            ema_inplace(self.cluster_size, cluster_size, self.decay)
+
             embed_sum = flatten.t() @ embed_onehot
+            self.all_reduce_fn(embed_sum)
+
             ema_inplace(self.embed_avg, embed_sum.t(), self.decay)
             cluster_size = laplace_smoothing(self.cluster_size, self.codebook_size, self.eps) * self.cluster_size.sum()
             embed_normalized = self.embed_avg / cluster_size.unsqueeze(1)
@@ -169,7 +187,9 @@ class CosineSimCodebook(nn.Module):
         kmeans_iters = 10,
         decay = 0.8,
         eps = 1e-5,
-        threshold_ema_dead_code = 2
+        threshold_ema_dead_code = 2,
+        use_ddp = False,
+        learnable_codebook = False
     ):
         super().__init__()
         self.decay = decay
@@ -184,9 +204,15 @@ class CosineSimCodebook(nn.Module):
         self.eps = eps
         self.threshold_ema_dead_code = threshold_ema_dead_code
 
+        self.all_reduce_fn = distributed.all_reduce if use_ddp else noop
         self.register_buffer('initted', torch.Tensor([not kmeans_init]))
         self.register_buffer('cluster_size', torch.zeros(codebook_size))
-        self.register_buffer('embed', embed)
+
+        self.learnable_codebook = learnable_codebook
+        if learnable_codebook:
+            self.embed = nn.Parameter(embed)
+        else:
+            self.register_buffer('embed', embed)
 
     @torch.jit.ignore
     def init_embed_(self, data):
@@ -218,6 +244,7 @@ class CosineSimCodebook(nn.Module):
         batch_samples = rearrange(batch_samples, '... d -> (...) d')
         self.replace(batch_samples, mask = expired_codes)
 
+    @autocast(enabled = False)
     def forward(self, x):
         shape, dtype = x.shape, x.dtype
         flatten = rearrange(x, '... d -> (...) d')
@@ -225,7 +252,9 @@ class CosineSimCodebook(nn.Module):
 
         self.init_embed_(flatten)
 
-        embed = l2norm(self.embed)
+        embed = self.embed if not self.learnable_codebook else self.embed.detach()
+        embed = l2norm(embed)
+
         dist = flatten @ embed.t()
         embed_ind = dist.max(dim = -1).indices
         embed_onehot = F.one_hot(embed_ind, self.codebook_size).type(dtype)
@@ -235,12 +264,16 @@ class CosineSimCodebook(nn.Module):
 
         if self.training:
             bins = embed_onehot.sum(0)
+            self.all_reduce_fn(bins)
+
             ema_inplace(self.cluster_size, bins, self.decay)
 
             zero_mask = (bins == 0)
             bins = bins.masked_fill(zero_mask, 1.)
 
             embed_sum = flatten.t() @ embed_onehot
+            self.all_reduce_fn(embed_sum)
+
             embed_normalized = (embed_sum / bins.unsqueeze(0)).t()
             embed_normalized = l2norm(embed_normalized)
             embed_normalized = torch.where(zero_mask[..., None], embed,
@@ -272,7 +305,7 @@ class VectorQuantize(nn.Module):
         orthogonal_reg_weight = 0.,
         orthogonal_reg_active_codes_only = False,
         orthogonal_reg_max_codes = None,
-        quantize_full_precision = False
+        sync_codebook = False
     ):
         super().__init__()
         n_embed = default(n_embed, codebook_size)
@@ -287,6 +320,7 @@ class VectorQuantize(nn.Module):
         self.eps = eps
         self.commitment_weight = default(commitment_weight, commitment)
 
+        has_codebook_orthogonal_loss = orthogonal_reg_weight > 0
         self.orthogonal_reg_weight = orthogonal_reg_weight
         self.orthogonal_reg_active_codes_only = orthogonal_reg_active_codes_only
         self.orthogonal_reg_max_codes = orthogonal_reg_max_codes
@@ -301,7 +335,9 @@ class VectorQuantize(nn.Module):
             kmeans_iters = kmeans_iters,
             decay = decay,
             eps = eps,
-            threshold_ema_dead_code = threshold_ema_dead_code
+            threshold_ema_dead_code = threshold_ema_dead_code,
+            use_ddp = sync_codebook,
+            learnable_codebook = has_codebook_orthogonal_loss
         )
 
         self.codebook_size = codebook_size
@@ -309,9 +345,6 @@ class VectorQuantize(nn.Module):
         self.accept_image_fmap = accept_image_fmap
         self.channel_last = channel_last
 
-        # when quantizing, whether to turn off mixed precision
-        self.quantize_full_precision = quantize_full_precision
-
     @property
     def codebook(self):
         return self._codebook.embed
@@ -330,10 +363,7 @@ class VectorQuantize(nn.Module):
 
         x = self.project_in(x)
 
-        context = null_context() if not self.quantize_full_precision else autocast(enabled = False)
-
-        with context:
-            quantize, embed_ind = self._codebook(x)
+        quantize, embed_ind = self._codebook(x)
 
         if self.training:
             quantize = x + (quantize - x).detach()
